@@ -7,7 +7,7 @@
 #' 
 #' 'LaBB-CAT' is a web-based language corpus management system and this
 #' package provides access to data stored in a 'LaBB-CAT' instance.
-#' You must have at least version 20190425.1121 of 'LaBB-CAT' to use
+#' You must have at least version 20191022.1827 of 'LaBB-CAT' to use
 #' this package.
 #' 
 #' @docType package
@@ -37,10 +37,8 @@ NULL
 ### Internal variables:
 
 ## minimum version of LaBB-CAT required:
-.min.labbcat.version <- "20190412.1154"
+.min.labbcat.version <- "20191022.1827"
 
-## HTTP request timeout
-.request.timeout <- 10
 
 ## prompt for password in RStudio, falling back to terminal if we're not in RStudio
 get.hidden.input <- function(prompt) {
@@ -86,7 +84,7 @@ store.get <- function(labbcat.url, call, parameters = NULL) {
     url <- enc(url)
     
     ## attempt the request
-    resp <- httr::GET(url, httr::timeout(.request.timeout))
+    resp <- httr::GET(url, httr::timeout(getOption("nzilbb.labbcat.timeout", default=10)))
     ## check we don't need credentials
     if (httr::status_code(resp) == 401 && interactive()) {
         ## ask for username and password
@@ -132,7 +130,7 @@ http.get <- function(labbcat.url, path, parameters = NULL, content.type = "appli
     url <- enc(url)
     
     ## attempt the request
-    resp <- httr::GET(url, httr::timeout(.request.timeout), httr::add_headers("Accepts" = content.type))
+    resp <- httr::GET(url, httr::timeout(getOption("nzilbb.labbcat.timeout", default=10)), httr::add_headers("Accepts" = content.type))
     ## check we don't need credentials
     if (httr::status_code(resp) == 401 && interactive()) {
         ## ask for username and password
@@ -166,16 +164,16 @@ http.get <- function(labbcat.url, path, parameters = NULL, content.type = "appli
 
 ## make an HTTP POST request, asking for credentials if required
 http.post <- function(labbcat.url, path, parameters, file.name) {
+    
     ## ensure labbcat base URL has a trailing slash
     if (!grepl("/$", labbcat.url)) labbcat.url <- paste(labbcat.url, "/", sep="")
 
     ## build request URL
     url <- paste(labbcat.url, path, sep="")
-    
     ## attempt the request
-    resp <- httr::POST(url,
+    resp <- httr::POST(url, 
                        httr::write_disk(file.name, overwrite=TRUE),
-                       httr::timeout(.request.timeout),
+                       httr::timeout(getOption("nzilbb.labbcat.timeout", default=10)),
                        body = parameters, encode = "form")
     ## check we don't need credentials
     if (httr::status_code(resp) == 401 && interactive()) {
@@ -219,7 +217,7 @@ http.post.multipart <- function(labbcat.url, path, parameters, file.name) {
     ## attempt the request
     resp <- httr::POST(url,
                        httr::write_disk(file.name, overwrite=TRUE),
-                       httr::timeout(.request.timeout),
+                       httr::timeout(getOption("nzilbb.labbcat.timeout", default=10)),
                        body = parameters, encode = "multipart")
     ## check we don't need credentials
     if (httr::status_code(resp) == 401 && interactive()) {
@@ -290,7 +288,7 @@ labbcatCredentials <- function(labbcat.url, username, password) {
     
     version.check.url <- paste(labbcat.url, "store?call=", sep="")
     authorization <- httr::authenticate(username, password)
-    resp <- httr::GET(version.check.url, authorization, httr::timeout(.request.timeout))
+    resp <- httr::GET(version.check.url, authorization, httr::timeout(getOption("nzilbb.labbcat.timeout", default=10)))
 
     if (httr::status_code(resp) != 200) { # 200 = OK
         if (httr::status_code(resp) == 401) {
@@ -303,7 +301,7 @@ labbcatCredentials <- function(labbcat.url, username, password) {
 
     ## do a second request
     ## - this seems to be required for credentials to 'take' in non-interactive mode
-    resp <- httr::GET(version.check.url, authorization, httr::timeout(.request.timeout))
+    resp <- httr::GET(version.check.url, authorization, httr::timeout(getOption("nzilbb.labbcat.timeout", default=10)))
     
     ## check the LaBB-CAT version
     resp.content <- httr::content(resp, as="text", encoding="UTF-8")
@@ -316,6 +314,26 @@ labbcatCredentials <- function(labbcat.url, username, password) {
     return(TRUE)    
 }
 
+#' Sets the timeout for request to the LaBB-CAT server in future function calls. The
+#' default timeout is 10 seconds.
+#'
+#' @param seconds The number of seconds before requests return with a timeout error.
+#' @return The request timeout in seconds
+#' @examples
+#' \dontrun{
+#' ## the request timeout
+#' labbcatTimeout(30)
+#' }
+#'
+#' @keywords connect username password timeout
+#' 
+labbcatTimeout <- function(seconds=NULL) {
+    if (!is.null(seconds)) { 
+        options(nzilbb.labbcat.timeout=seconds)
+    }
+    return(getOption("nzilbb.labbcat.timeout", default=10))
+}
+
 #' Gets the store's ID.
 #' 
 #' The store's ID - i.e. the ID of the 'LaBB-CAT' instance.
@@ -652,7 +670,8 @@ getGraphIdsWithParticipant <- function(labbcat.url, id) {
 #' @param pageLength The maximum number of IDs to return, or null to return all
 #' @param pageNumber The zero-based page number to return, or null to return the first page
 #' @param order An expression that determines the order the graphs are
-#' listed in
+#' listed in - if specified, this must include the keyword 'ASC' for ascending or 'DESC'
+#' for descending order.
 #' @return A list of graph IDs (i.e. transcript names)
 #' 
 #' @examples 
@@ -675,7 +694,7 @@ getGraphIdsWithParticipant <- function(labbcat.url, id) {
 #' ## in word-count order 
 #' transcripts <- getMatchingGraphIds(
 #'         labbcat.url, "my('corpus').label = 'QB' AND 'QB247_Jacqui' IN labels('who')", 1, 1,
-#'         "my('transcript_word_count').label")
+#'         "my('transcript_word_count').label ASC")
 #' }
 #' 
 #' @keywords graph transcript expression
@@ -931,6 +950,7 @@ getMedia <- function(labbcat.url, id, trackSuffix = "", mimeType = "audio/wav")
 #'     integer, then the result is a mono file with the given sample rate.
 #' @param no.progress Optionally suppress the progress bar when
 #'     multiple fragments are  specified - TRUE for no progress bar.
+#' @param path Optional path to directory where tge files should be saved.
 #' @return The name of the file, which is saved in the current
 #'     directory, or a list of names of files, if multiple
 #'     id's/start's/end's were specified 
@@ -945,7 +965,7 @@ getMedia <- function(labbcat.url, id, trackSuffix = "", mimeType = "audio/wav")
 #' labbcat.url <- "https://labbcat.canterbury.ac.nz/demo/"
 #' 
 #' ## Get the 5 seconds starting from 10s after the beginning of a recording
-#' wav.file <- getSoundFragments(labbcat.url, "AP2505_Nelson.eaf", 10.0, 15.0)
+#' wav.file <- getSoundFragments(labbcat.url, "AP2505_Nelson.eaf", 10.0, 15.0, path="samples")
 #' 
 #' ## Get the 5 seconds starting from 10s as a mono 22kHz file
 #' wav.file <- getSoundFragments(labbcat.url, "AP2505_Nelson.eaf", 10.0, 15.0, 22050)
@@ -956,19 +976,22 @@ getMedia <- function(labbcat.url, id, trackSuffix = "", mimeType = "audio/wav")
 #' ## Get a list of fragments
 #' wav.files <- getSoundFragments(labbcat.url, results$Transcript, results$Line, results$LineEnd)
 #' 
-#' ## Get a list of fragments with no prgress bar
+#' ## Get a list of fragments with no progress bar
 #' wav.file <- getSoundFragments(
 #'               labbcat.url, results$Transcript, results$Line, results$LineEnd, no.progress=TRUE)
 #' }
 #' @keywords sample sound fragment wav
 #' 
-getSoundFragments <- function(labbcat.url, id, start, end, sampleRate = NULL, no.progress=FALSE) {
-    if (length(id) == 1) { ## one fragment
-        dir <- ""
-    } else { ## multiple fragments
+getSoundFragments <- function(labbcat.url, id, start, end, sampleRate = NULL, no.progress=FALSE, path="") {
+    
+    dir = path
+    if (length(id) > 1) { ## multiple fragments
         ## save fragments into their own directory
-        dir <- "fragments"
-        if (file.exists(dir)) {
+        if (stringr::str_length(dir) == 0) dir <- "fragments"
+    }
+    if (stringr::str_length(dir) > 0) { ## directory is specified
+        ## if it wasn't explicitly specified...
+        if (file.exists(dir) && stringr::str_length(path) == 0) { 
             ## ensure it's a new directory by adding a number
             n <- 1
             new.dir = paste(dir,"(",n,")", sep="")
@@ -978,9 +1001,11 @@ getSoundFragments <- function(labbcat.url, id, start, end, sampleRate = NULL, no
             } # next try
             dir <- new.dir
         }
-        dir.create(dir)
-        ## add trailing slash
-        dir <- paste(dir, .Platform$file.sep, sep="")
+        if (!file.exists(dir)) dir.create(dir)
+        ## add trailing slash if there isn't one
+        if (!grepl(paste("\\", .Platform$file.sep, "$", sep=""), dir)) {
+            dir <- paste(dir, .Platform$file.sep, sep="")
+        }
     }
 
     pb <- NULL
@@ -1033,6 +1058,132 @@ getSoundFragments <- function(labbcat.url, id, start, end, sampleRate = NULL, no
     return(file.names)   
 }
 
+#' Gets fragments of annotation graphs (transcripts) from 'LaBB-CAT', 
+#' converted to a given format (by default, Praat TextGrid)
+#'
+#' @param labbcat.url URL to the LaBB-CAT instance
+#' @param id The graph ID (transcript name) of the sound recording, or
+#'     a vector of graph IDs. 
+#' @param start The start time in seconds, or a vector of start times.
+#' @param end The end time in seconds, or a vector of end times.
+#' @param layerId A vector of layer IDs.
+#' @param mimeType Optional content-type - currently only "text/praat-textgrid"
+#'     is supported.
+#' @param no.progress Optionally suppress the progress bar when
+#'     multiple fragments are  specified - TRUE for no progress bar.
+#' @param path Optional path to directory where tge files should be saved.
+#' @return The name of the file, which is saved in the current
+#'     directory, or a list of names of files, if multiple
+#'     id's/start's/end's were specified 
+#'
+#' If a list of files is returned, they are in the order that they
+#'     were returned by the server, which *should* be the order that
+#'     they were specified in the id/start/end lists.
+#' 
+#' @examples
+#' \dontrun{
+#' ## define the LaBB-CAT URL
+#' labbcat.url <- "https://labbcat.canterbury.ac.nz/demo/"
+#' 
+#' ## Get the 5 seconds starting from 10s after the beginning of a recording
+#' textgrid.file <- getFragments(labbcat.url, "AP2505_Nelson.eaf", 10.0, 15.0,
+#'     c("transcript", "phonemes"), path="samples") 
+#' 
+#' ## Load some search results previously exported from LaBB-CAT
+#' results <- read.csv("results.csv", header=T)
+#' 
+#' ## Get a list of fragment TextGrids, including the utterances, transcript, and phonemes layers
+#' textgrid.files <- getFragments(
+#'     labbcat.url, results$Transcript, results$Line, results$LineEnd,
+#'     c("utterances", "transcript", "phonemes"))
+#' 
+#' ## Get a list of fragment TextGrids with no progress bar
+#' textgrid.files <- getFragments(
+#'     labbcat.url, results$Transcript, results$Line, results$LineEnd, no.progress=TRUE)
+#' }
+#' @keywords sample sound fragment wav
+#' 
+getFragments <- function(labbcat.url, id, start, end, layerId, mimeType = "text/praat-textgrid", no.progress=FALSE, path="") {
+
+    dir = path
+    if (length(id) > 1) { ## multiple fragments
+        ## save fragments into their own directory
+        if (stringr::str_length(dir) == 0) dir <- "fragments"
+    }
+    if (stringr::str_length(dir) > 0) { ## directory is specified
+        ## if it wasn't explicitly specified...
+        if (file.exists(dir) && stringr::str_length(path) == 0) { 
+            ## ensure it's a new directory by adding a number
+            n <- 1
+            new.dir = paste(dir,"(",n,")", sep="")
+            while (file.exists(new.dir)) {
+                n <- n + 1
+                new.dir = paste(dir,"(",n,")", sep="")
+            } # next try
+            dir <- new.dir
+        }
+        if (!file.exists(dir)) dir.create(dir)
+        ## add trailing slash if there isn't one
+        if (!grepl(paste("\\", .Platform$file.sep, "$", sep=""), dir)) {
+            dir <- paste(dir, .Platform$file.sep, sep="")
+        }
+    }
+    
+    pb <- NULL
+    if (!no.progress && length(id) > 1) {
+        pb <- txtProgressBar(min = 0, max = length(id), style = 3)        
+    }
+
+    ## create a list of repeated layerId parameters
+    layerParameters <- list()
+    mapply(function(l) { layerParameters <<- c(layerParameters, list(layerId=l)) }, layerId)
+
+    ## loop throug each triple, getting fragments individually
+    ## (we could actually pass the lot to LaBB-CAT in one go and get a ZIP file back
+    ##  but then we can't be sure the results contain a row for every fragment specified
+    ##  and we can't display a progress bar)
+    file.names = c()
+    r <- 1
+    for (graph.id in id) {
+        parameters <- list(mimeType=mimeType, id=graph.id, start=start[r], end=end[r])
+        ## add layerId parameters
+        parameters <- c(parameters, layerParameters)
+        file.name <- paste(dir, stringr::str_replace(graph.id, "\\.[^.]+$",""), "__", start[r], "-", end[r], ".TextGrid", sep="")
+        
+        tryCatch({
+            resp <- http.post(labbcat.url, "convertfragment", parameters, file.name)
+            if (httr::status_code(resp) != 200) { # 200 = OK
+                print(paste("ERROR: ", httr::http_status(resp)$message))
+                if (httr::status_code(resp) != 404) { # 404 means the audio wasn't on the server
+                    ## some other error occurred so print what we got from the server
+                    print(readLines(file.name))
+                }
+                file.remove(file.name)
+                file.name <<- NULL
+            } else {
+                content.disposition <- as.character(httr::headers(resp)["content-disposition"])
+                content.disposition.parts <- strsplit(content.disposition, "=")
+                if (length(content.disposition.parts[[1]]) > 1
+                    && file.name != content.disposition.parts[[1]][2]) {
+                    ## file name is specified, so use it
+                    final.file.name <- paste(dir, content.disposition.parts[[1]][2], sep="")
+                    file.rename(file.name, final.file.name)
+                    file.name <- final.file.name
+                }
+            }
+        }, error = function(e) {
+            print(paste("ERROR:", e))
+            file.name <<- NULL
+        })
+        file.names <- append(file.names, file.name)
+        
+        if (!is.null(pb)) setTxtProgressBar(pb, r)
+        r <- r+1
+    } ## next row
+    if (!is.null(pb)) close(pb)
+    return(file.names)   
+}
+
 #' Gets labels of annotations on a given layer, identified by given annotation IDs.
 #'
 #' @param labbcat.url URL to the LaBB-CAT instance
